"""
xmppwb.bridge
~~~~~~~~~~~~~

This module implements the main bridging functionality.

:copyright: (c) 2016 by saqura.
:license: MIT, see LICENSE for more details.
"""
import json
import logging
import os
import ssl
import aiohttp
import aiohttp.web

from xmppwb.xmpp import XMPPBridgeBot


class XMPPWebhookBridge:
    """This is the central component: It initializes and connects the
    :class:`XMPPBridgeBot` with the webhook handling part. The webhook part
    consists of an HTTP server listening for incoming webhooks (POST requests),
    and an HTTP client part for sending outgoing webhooks (POST requests).
    """
    def __init__(self, cfg, loop):
        self.loop = loop
        # List of bridges
        self.bridges = list()
        # Mapping of MUC-JID -> Nickname
        self.mucs = dict()
        # Mapping of MUC-JID -> Password
        self.muc_passwords = dict()

        try:
            # Get the optional XMPP address (host, port) if specified
            xmpp_address = tuple()
            if 'host' in cfg['xmpp']:
                xmpp_address = (cfg['xmpp']['host'], cfg['xmpp']['port'])
            # Parse the MUC definitions
            self.get_mucs(cfg)
            if len(self.mucs) == 0:
                logging.info("No MUCs defined.")
        except KeyError:
            raise InvalidConfigError

        # Parse the bridges from the config file
        need_incoming_webhooks = False
        for bridge_cfg in cfg['bridges']:
            bridge = SingleBridge(bridge_cfg, self)
            self.bridges.append(bridge)
            if bridge.has_incoming_webhooks():
                need_incoming_webhooks = True

        # Initialize XMPP client
        self.xmpp_client = XMPPBridgeBot(cfg['xmpp']['jid'],
                                         cfg['xmpp']['password'],
                                         self)
        self.xmpp_client.connect(address=xmpp_address)

        # Initialize HTTP server if needed
        if not need_incoming_webhooks:
            self.http_server = None
            logging.info("No incoming webhooks defined.")
        elif 'incoming_webhook_listener' in cfg:
            bind_address = cfg['incoming_webhook_listener']['bind_address']
            port = cfg['incoming_webhook_listener']['port']
            self.http_app = aiohttp.web.Application(loop=loop)
            self.http_app.router.add_route('POST',
                                           '/',
                                           self.handle_incoming_webhook)
            self.http_handler = self.http_app.make_handler()
            http_create_server = loop.create_server(
                self.http_handler,
                bind_address,
                port)
            self.http_server = loop.run_until_complete(http_create_server)
            logging.info("Listening for incoming webhooks on "
                         "http://{}:{}/".format(bind_address, port))
        else:
            self.http_server = None
            logging.warn("No 'incoming_webhook_listener' in the config even "
                         "though incoming webhooks are defined. Ignoring all "
                         "incoming webhooks.")

        if not self.http_server:
            logging.info("Not listening for incoming webhooks.")

    def process(self):
        self.loop.run_forever()

    async def send_outgoing_webhook(self, outgoing_webhook, msg):
        """This coroutine handles outgoing webhooks: It relays the messages
        received from XMPP and triggers external webhooks.
        """
        from_jid = msg['from']
        username = str(from_jid)
        if 'override_username' in outgoing_webhook:
            if msg['type'] == 'groupchat':
                username = outgoing_webhook['override_username'].format(
                    bare_jid=from_jid.bare,
                    full_jid=from_jid.full,
                    local_jid=from_jid.local,
                    nick=from_jid.resource,
                    jid=from_jid.full)
            else:
                username = outgoing_webhook['override_username'].format(
                    bare_jid=from_jid.bare,
                    full_jid=from_jid.full,
                    local_jid=from_jid.local,
                    nick=from_jid.local,
                    jid=from_jid.bare)

        message = msg['body']
        if 'message_template' in outgoing_webhook:
            message = outgoing_webhook['message_template'].format(
                msg=message)

        payload = {
            'text': message,
            'username': username
        }

        if 'override_channel' in outgoing_webhook:
            payload['channel'] = outgoing_webhook['override_channel']

        # Attachment formatting is useful for integrating with RocketChat.
        if ('use_attachment_formatting' in outgoing_webhook and
                outgoing_webhook['use_attachment_formatting']):
            payload_attachment = {
                'title': "From: {}".format(username),
                'text': message
            }
            if 'attachment_link' in outgoing_webhook:
                payload_attachment['title_link'] = \
                                            outgoing_webhook['attachment_link']
            payload = {
                'attachments': [payload_attachment]
            }

        logging.debug("<-- Sending outgoing webhook. (from '{}')".format(
                                                            from_jid))
        request = await outgoing_webhook['session'].post(
            outgoing_webhook['url'],
            data=json.dumps(payload),
            headers={'content-type': 'application/json'})
        await request.release()
        return

    async def handle_incoming_webhook(self, request):
        """This coroutine handles incoming webhooks: It receives incoming
        webhooks and relays the messages to XMPP."""
        if request.content_type == 'application/json':
            payload = await request.json()
            # print(payload)
        else:
            # TODO: Handle other content types
            payload = await request.post()

        # Disgard empty messages
        if payload['text'] == "":
            return aiohttp.web.Response()

        token = payload['token']
        logging.debug("--> Handling incoming request from token "
                      "'{}'...".format(token))
        username = payload['user_name']
        msg = payload['text']

        for bridge in self.bridges:
            bridge.handle_incoming_webhook(token, username, msg)

        return aiohttp.web.Response()

    def get_mucs(self, cfg):
        """Reads the MUC definitions from the config file."""
        if 'mucs' not in cfg['xmpp']:
            return

        for muc in cfg['xmpp']['mucs']:
            jid = muc['jid']
            nickname = muc['nickname']
            self.mucs[jid] = nickname
            if 'password' in muc:
                self.muc_passwords[jid] = muc['password']

    def close(self):
        """Closes all open connections, servers and handlers. This is used
        when exiting the bridge.
        """
        if self.http_server:
            logging.info("Closing HTTP server...")
            self.http_server.close()
            self.loop.run_until_complete(self.http_server.wait_closed())
            self.loop.run_until_complete(self.http_handler.
                                         finish_connections(1.0))
            self.loop.run_until_complete(self.http_app.finish())
            logging.info("Closed HTTP server..")

        logging.info("Closing HTTP client sessions...")
        for bridge in self.bridges:
            for webhook in bridge.outgoing_webhooks:
                webhook['session'].close()
        logging.info("Closed HTTP client sessions..")
        logging.info("Disconnecting from XMPP...")
        self.xmpp_client.disconnect()
        logging.info("Disconnected from XMPP.")


class InvalidConfigError(Exception):
    """Raised when the config file is invalid."""
    pass


class SingleBridge:
    def __init__(self, bridge_cfg, main_bridge):
        """Parses a bridge section of the config file and creates a new
        bridge.
        """
        self.main_bridge = main_bridge
        self.xmpp_muc_endpoints = list()
        self.xmpp_normal_endpoints = list()
        self.xmpp_relay_all_normal = False

        # List of tokens of incoming webhooks for this bridge
        self.incoming_webhooks = list()

        self.outgoing_webhooks = list()

        self._parse_xmpp_endpoints(bridge_cfg)
        self._parse_incoming_webhooks(bridge_cfg)
        self._parse_outgoing_webhooks(bridge_cfg)

    def has_incoming_webhooks(self):
        """Returns True if this bridge contains incoming webhooks."""
        return (len(self.incoming_webhooks) != 0)

    def handle_incoming_webhook(self, token, username, msg):
        """Handles an incoming webhook with the given token, username
        and message.
        """
        if token not in self.incoming_webhooks:
            # This webhook is not handled by this bridge.
            return

        self.send_to_all_xmpp_endpoints(username, msg)

    def send_to_all_xmpp_endpoints(self, username, msg, skip=list()):
        """Send the given message from the given user to all XMPP endpoints
        of this bridge, except for the JIDs in the `skip`-list.
        """
        msg = "{}: {}".format(username, msg)
        for xmpp_normal_jid in self.xmpp_normal_endpoints:
            if xmpp_normal_jid in skip:
                continue

            logging.debug("<-- Sending a normal chat message to XMPP.")
            self.main_bridge.xmpp_client.send_message(
                mto=xmpp_normal_jid,
                mbody=msg,
                mtype='chat',
                mnick=username)

        for xmpp_muc_jid in self.xmpp_muc_endpoints:
            if xmpp_muc_jid in skip:
                continue

            logging.debug("<-- Sending a MUC chat message to XMPP.")
            self.main_bridge.xmpp_client.send_message(
                mto=xmpp_muc_jid,
                mbody=msg,
                mtype='groupchat',
                mnick=username)

    async def handle_incoming_xmpp(self, msg):
        """Handles an incoming XMPP message, from either a normal chat or
        a MUC."""
        # Outgoing webhooks to trigger
        out_webhooks = list()
        from_jid = msg['from']

        if msg['type'] in ('chat', 'normal'):
            if from_jid.bare in self.xmpp_normal_endpoints:
                out_webhooks = self.outgoing_webhooks

                # Relay this message to the other XMPP endpoints of this bridge
                self.send_to_all_xmpp_endpoints(from_jid.local,
                                                msg['body'],
                                                skip=[from_jid.bare])

            elif self.xmpp_relay_all_normal:
                out_webhooks = self.outgoing_webhooks

        elif msg['type'] == 'groupchat':
            # TODO: Handle nickname of private message in MUCs.
            if from_jid.resource == self.main_bridge.mucs[from_jid.bare]:
                # Don't relay messages from ourselves.
                return
            elif from_jid.bare in self.xmpp_muc_endpoints:
                out_webhooks = self.outgoing_webhooks

                # Relay this message to the other XMPP endpoints of this bridge
                self.send_to_all_xmpp_endpoints(from_jid.resource,
                                                msg['body'],
                                                skip=[from_jid.bare])

        else:
            # Only handle normal chats and MUCs.
            return

        # Forward the messages to the outgoing webhooks
        for outgoing_webhook in out_webhooks:
            await self.main_bridge.send_outgoing_webhook(outgoing_webhook, msg)

    def _parse_incoming_webhooks(self, bridge_cfg):
        """Parses the `incoming_webhooks` from this bridge's config file
        section."""
        if 'incoming_webhooks' not in bridge_cfg:
            # No incoming webhooks in this bridge.
            return

        incoming_webhooks = bridge_cfg['incoming_webhooks']
        for incoming_webhook in incoming_webhooks:
            if 'token' not in incoming_webhook:
                raise InvalidConfigError("Invalid config file: "
                                         "'token' missing from outgoing "
                                         "webhook definition.")
            token = incoming_webhook['token']
            self.incoming_webhooks.append(token)

    def _parse_outgoing_webhooks(self, bridge_cfg):
        """Parses the `outgoing webhooks` from this bridge's config file
        section.

        This also sets up the HTTP client session for each webhook."""
        if 'outgoing_webhooks' not in bridge_cfg:
            # No outgoing webhooks in this bridge.
            return

        outgoing_webhooks = bridge_cfg['outgoing_webhooks']

        for outgoing_webhook in outgoing_webhooks:
            if 'url' not in outgoing_webhook:
                raise InvalidConfigError("Error in config file: "
                                         "'url' is missing from an "
                                         "outgoing webhook definition.")

            # Set up SSL context for certificate pinning.
            if 'cafile' in outgoing_webhook:
                cafile = os.path.abspath(outgoing_webhook['cafile'])
                sslcontext = ssl.create_default_context(cafile=cafile)
                conn = aiohttp.TCPConnector(ssl_context=sslcontext)
                session = aiohttp.ClientSession(loop=self.main_bridge.loop,
                                                connector=conn)
            else:
                session = aiohttp.ClientSession(loop=self.main_bridge.loop)
            # TODO: Handle ConnectionRefusedError.
            outgoing_webhook['session'] = session

            self.outgoing_webhooks.append(outgoing_webhook)

    def _parse_xmpp_endpoints(self, bridge_cfg):
        """Parses the `xmpp_endpoints` from this bridge's config file
        section."""
        if 'xmpp_endpoints' not in bridge_cfg:
            raise InvalidConfigError("Error in config file: "
                                     "'xmpp_endpoints' section is missing "
                                     "from a bridge definition.")

        xmpp_endpoints = bridge_cfg['xmpp_endpoints']
        for xmpp_endpoint in xmpp_endpoints:
            # Determine whether the JID corresponds to a MUC or a
            # normal chat:
            if 'muc' in xmpp_endpoint:
                muc_endpoint = xmpp_endpoint['muc']
                if muc_endpoint not in self.main_bridge.mucs:
                    raise InvalidConfigError(
                        "Error in config file: XMPP MUC '{}' was not "
                        "defined in the xmpp.mucs section.".format(
                            muc_endpoint))
                self.xmpp_muc_endpoints.append(muc_endpoint)

            elif 'normal' in xmpp_endpoint:
                normal_endpoint = xmpp_endpoint['normal']
                self.xmpp_normal_endpoints.append(normal_endpoint)

            elif 'relay_all_normal' in xmpp_endpoint:
                if xmpp_endpoint['relay_all_normal'] is True:
                    self.xmpp_relay_all_normal = True

            else:
                raise InvalidConfigError("Error in config file: "
                                         "'xmpp_endpoints' section contains "
                                         "invalid entry. Must be either of "
                                         "type 'muc', 'normal' or "
                                         "'relay_all_normal'.")



"""
xmppwb.xmpp
~~~~~~~~~~~

This module implements the XMPP specific parts of the bridge.

:copyright: (c) 2016 by saqura.
:license: MIT, see LICENSE for more details.
"""
import logging
from slixmpp import ClientXMPP


class XMPPBridgeBot(ClientXMPP):
    """The XMPP part of the bridge. It is a bot that connects to all specified
    MUCs, listens to incoming messages (from both MUCs and normal chats) and
    sends messages.
    """
    def __init__(self, jid, password, main_bridge):
        ClientXMPP.__init__(self, jid, password)

        self.main_bridge = main_bridge

        self.add_event_handler("session_start", self.session_started)
        self.add_event_handler("message", self.message_received)
        self.add_event_handler("connection_failed", self.connection_failed)
        self.add_event_handler("failed_auth", self.auth_failed)

        self.register_plugin('xep_0030')  # Service Discovery
        self.register_plugin('xep_0045')  # Multi-User Chat
        self.register_plugin('xep_0199')  # XMPP Ping

    def session_started(self, event):
        """This sets up the XMPP bot once successfully connected. It connects
        to all specified MUCs.
        """
        self.send_presence()
        self.get_roster()

        for muc, nickname in self.main_bridge.mucs.items():
            logging.debug("Joining MUC '{}' using nickname '{}'.".format(
                                                                muc, nickname))
            if muc in self.main_bridge.muc_passwords:
                self.plugin['xep_0045'].joinMUC(
                    muc,
                    nickname,
                    wait=True,
                    password=self.main_bridge.muc_passwords[muc])
            else:
                self.plugin['xep_0045'].joinMUC(muc,
                                                nickname,
                                                wait=True)

        logging.info("Connected to XMPP.")

    async def message_received(self, msg):
        """This coroutine is triggered whenever a message (both normal or from
        a MUC) is received. It relays the message to the bridge.
        """
        logging.debug("--> Received message from XMPP by {}: {}".format(
            msg['from'], msg['body']))

        for bridge in self.main_bridge.bridges:
            await bridge.handle_incoming_xmpp(msg)

    async def connection_failed(self, error):
        """This coroutine is triggered when the connection to the XMPP server
        failed.
        """
        logging.error("Connection to XMPP failed.")

    async def auth_failed(self, error):
        """This coroutine is triggered when the XMPP server has rejected the
        login credentials.
        """
        logging.error("Authetication with XMPP failed.")



"""
xmppwb.core
~~~~~~~~~~~

This module is mainly used as an entrypoint to set everything up.

:copyright: (c) 2016 by saqura.
:license: MIT, see LICENSE for more details.
"""
import argparse
import asyncio
import logging
import os
import sys
import yaml

from xmppwb.bridge import XMPPWebhookBridge, InvalidConfigError
from xmppwb import __version__


def main():
    """Main entry point.

    Gathers the command line arguments, reads the config and starts the bridge.
    """
    parser = argparse.ArgumentParser(
        description="A bot that bridges XMPP (chats and MUCs) with webhooks, "
        "thus making it possible to interact with services outside the XMPP "
        "world.")
    parser.add_argument("-c", "--config", help="set the config file",
                        required=True)
    parser.add_argument("-v", "--verbose", help="increase output verbosity",
                        action="store_true")
    parser.add_argument("-l", "--logfile", help="enable logging to a file")
    parser.add_argument("-d", "--debug", help="include debug output",
                        action="store_true")
    parser.add_argument("--version", help="show version and exit",
                        action="version", version=__version__)
    args = parser.parse_args()

    loop = asyncio.get_event_loop()

    log_config = {
        'format': '%(asctime)s %(levelname)-8s %(message)s',
        'datefmt': '%d-%H:%M:%S'
    }

    if args.debug:
        loop.set_debug(True)
        args.verbose = True

    if args.verbose:
        log_config['level'] = logging.DEBUG
    else:
        log_config['level'] = logging.INFO

    if args.logfile:
        log_config['filename'] = args.logfile


    logging.getLogger('slixmpp').setLevel(logging.WARNING)
    logging.getLogger('aiohttp').setLevel(logging.WARNING)
    # logger = logging.getLogger('xmppwb')
    logging.basicConfig(**log_config)

    logging.info("Starting xmppwb version {}".format(__version__))
    config_filepath = os.path.abspath(args.config)
    logging.info("Using config file {}".format(config_filepath))
    try:
        with open(config_filepath, 'r') as config_file:
            cfg = yaml.load(config_file)
    except FileNotFoundError:
        logging.error("Config file not found. Exiting...")
        sys.exit(1)

    try:
        bridge = XMPPWebhookBridge(cfg, loop)
    except InvalidConfigError:
        logging.exception("Invalid config file.")
        sys.exit(1)

    try:
        bridge.process()
    except KeyboardInterrupt:
        print("Exiting... (keyboard interrupt)")
    finally:
        bridge.close()
    loop.close()
    logging.info("xmppwb exited.")


if __name__ == '__main__':
    main()



"""
xmppwb._version
~~~~~~~~~~~~~~~

This module contains the xmppwb version.

:copyright: (c) 2016 by saqura.
:license: MIT, see LICENSE for more details.
"""

# The version is in a seperate file so that we can easily extract it from
# setup.py without importing the whole package.
# See: http://python-packaging-user-guide.readthedocs.org/en/latest/single_source_version/

__version__ = '0.4.0'



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the version
with open(path.join(here, 'xmppwb', 'version.py')) as f:
    vf = {}
    exec(f.read(), vf)
    version = vf['__version__']

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='xmppwb',
    version=version,
    description='XMPP Webhook Bridge',
    long_description=long_description,
    url='https://github.com/saqura/xmppwb',
    author='saqura',
    author_email='saqura@saqura.xyz',
    license='MIT',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Communications :: Chat',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
    ],
    keywords=['jabber', 'xmpp', 'bridge', 'bot', 'webhook', 'webhooks'],
    packages=find_packages(),
    install_requires=['aiohttp', 'pyyaml', 'slixmpp'],
    entry_points={
        'console_scripts': [
            'xmppwb=xmppwb.core:main',
        ],
    },
)



"""
xmppwb - XMPP Webhook Bridge
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A bot that bridges XMPP (chats and MUCs) with webhooks, thus making it
possible to interact with services outside the XMPP world. This can be used to
connect XMPP to other chat services that provide a webhook API to XMPP (for
example Rocket.Chat, Mattermost or Slack).

:copyright: (c) 2016 by saqura.
:license: MIT, see LICENSE for more details.
"""

from xmppwb.version import __version__
